# pyXLWriter: A library for generating Excel Spreadsheets
# Copyright (c) 2004 Evgeny Filatov <fufff@users.sourceforge.net>
# Copyright (c) 2002-2004 John McNamara (Perl Spreadsheet::WriteExcel)
#
# This library is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This library is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser
# General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this library; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#----------------------------------------------------------------------------
# This module was written/ported from PERL Spreadsheet::WriteExcel module
# The author of the PERL Spreadsheet::WriteExcel module is John McNamara
# <jmcnamara@cpan.org>
#----------------------------------------------------------------------------
# See the README.txt distributed with pyXLWriter for more details.

"""pyXLWriter2.Workbook

Used in conjunction with pyXLWriter2.

"""
__revision__ = """$Id: Workbook.py,v 1.28 2004/08/20 05:16:16 fufff Exp $"""

import re
from struct import pack, unpack
from BIFFWriter import BIFFWriter
from OLEWriter import OLEWriter
from OLEWriterBig import OLEWriterBig
from Format import Format
from Formula import Formula
from Worksheet import Worksheet

import tempfile


class Workbook(BIFFWriter):
    """Workbook - A writer class for Excel Workbooks."""

    def __init__(self, workfile, big=False):
        """Constructor."""
        BIFFWriter.__init__(self)
        tmp_format = Format()
        parser = Formula()
        self._file = workfile
        self._parser = parser
        self._1904 = 0
        self._activesheet = 0
        self._firstsheet = 0
        self._selected = 0
        self._xf_index = 16     # 15 style XF's and 1 cell XF.
        self._fileclosed = False
        self._biffsize = 0
        self._sheetname = "Sheet"
        self._tmp_format = tmp_format
        self._codepage = 0x04E4
        self._worksheets = []
        self._sheetnames = []
        self._formats = []
        self._palette = []
        self._big = big

        self._using_tmpfile    = True
        self._filehandle       = None
        self._temp_file        = None

        self._str_total        = 0
        self._str_unique       = 0
        self._str_table        = {}
        self._str_array        = []
        self._str_block_sizes  = []

        self._ext_ref_count    = 0
        self._ext_refs         = {}
        
        # Add the default format for hyperlinks and others...
        self._url_format = self.add_format(color="blue", underline=1)
        self._date_format = self.add_format()
        self._date_format.set_num_format("d mmmm yyy")
        self._time_format = self.add_format()
        self._time_format.set_num_format("HH:MM:SS")
        self._datetime_format = self.add_format()
        self._datetime_format.set_num_format("d mmmm yyy HH:MM:SS")
        
        # Set colour palette.
        self.set_palette_xl97()

        self._initialize()

    def _initialize(self):
        """Open a tmp file to store the majority of the Worksheet data. If this fails,
        for example due to write permissions, store the data in memory. This can be
        slow for large files.

        TODO: Move this and other methods shared with Worksheet up into BIFFWriter.
        """
        self._filehandle = tempfile.TemporaryFile()

    def _append(self, *data):
        """Store Worksheet data in memory using the base class _append() or to a
        temporary file, the default.
        """

        if self._using_tmpfile:
            data = "".join(data)
            if len(data) > self._limit:
                data = self._add_continue(data)
            self._filehandle.write(data)
            self._datasize += len(data)
        else:
            super(Workbook, self)._append(*data)

    def get_data(self):
        """Retrieves data from memory in one chunk, or from disk in $buffer
        sized chunks
        """
        buffer = 4096
        if self._data != "":
            tmp = self._data
            self._data = ""
            if self._using_tmpfile:
                self._filehandle.seek(0)
            return tmp
        if self._using_tmpfile:
            return self._filehandle.read(buffer)
        return ""

    def close(self):
        """Calls finalization methods and explicitly close the OLEwriter file
        handle.

        """
        try:
            if not self._fileclosed: # Prevent close() from being called twice.
                self._fileclosed = True
                self._store_workbook()
        finally:
            for sheet in self._worksheets:
                sheet._close_tmp()
            self._worksheets = [] # Simple fix memory leak :/

    #~ def __del__(self):
        #~ """Close the workbook if it hasn't already been explicitly closed."""
        #~ if not self._fileclosed:
            #~ self.close()

    def sheets(self):
        """An accessor for the _worksheets[] array

        Returns: a list of the worksheet objects in a workbook

        """
        return self._worksheets

    def worksheets(self):
        """An accessor for the _worksheets[] array.
        This method is now deprecated. Use the sheets() method instead.

        Returns: an array reference

        """
        return self._worksheets

    def add_worksheet(self, name="", encoding=0):
        """Add a new worksheet to the Excel workbook.
        TODO: Add accessor for self._sheetname for international Excel versions.

        Returns: reference to a worksheet object

        """
#         if encoding != 0:
#             limit = 62
#         else:
#             limit = 31
        limit = 31

        # Check that sheetname is <= 31 chars (Excel limit).
        if len(name) > limit:
            raise ValueError("Sheetname %s must be <= %d chars" % (name, limit))

        # Check that sheetname doesn't contain any invalid characters
        # r.match can't be used here because these characters can't be used
        # anywhere in the sheetname, not only at the beginning.
        # this codechunk also works for unicode strings. AM
        if re.search(r"[:*?/\\]", name):
            raise ValueError(r"Invalid Excel character [:*?/\\] in worksheet name: %s" % (name))

        index = len(self._worksheets)
        sheetname = self._sheetname

        if (name == ""):
            name = sheetname + str(index+1)

        # Check that the worksheet name doesn't already exist: a fatal Excel error.
        # The check must also exclude case insensitive matches.
        for tmp in self._worksheets:
            if name.lower() == tmp.get_name().lower():
                raise ValueError("Worksheet %s already exists" % (name))

        # Porters take note, the following scheme of passing references to Workbook
        # data (in the self._foo cases) instead of a reference to the Workbook
        # itself is a workaround to avoid circular references between Workbook and
        # Worksheet objects. Feel free to implement this in any way the suits your
        # language.
        #
        # AM: CHECKPOINT: get back here when Worksheet constructor will be checked and updated
        #     13th,14th chunks in diff of Workbook.pm's 1.01 and 2.03
        init_data = (
                     name,
                     self,
                     index,
                     #self._url_format,
                     self._parser,
                    )
        worksheet = Worksheet(*init_data)
        self._worksheets.append(worksheet)          # Store ref for iterator
        self._sheetnames.append(name)               # Store EXTERNSHEET names
        self._parser.set_ext_sheets(name, index)    # Store names in Formula.pm
        return worksheet
    
    def activate_sheet(self, sheet):
        """
            TODO:
        """
        sheet.select()
        self._activesheet = sheet.get_index()
    
    def set_first_sheet(self, sheet):
        """
            TODO:
        """
        self._firstsheet = sheet.get_index()
        
    def add_format(self, **properties):
        """Add a new format to the Excel workbook. This adds an XF record and
        a FONT record. Also, pass any properties to the Format::new().

        """
        format = Format(self._xf_index, **properties)
        self._xf_index += 1
        self._formats.append(format)
        return format

    def get_default_url_format(self):
        """Get default url format. It's used by worksheet.write_url() when 
        format is not defined.
        
        """
        return self._url_format
    
    def get_default_date_format(self):
        """Get default date format."""
        return self._date_format
        
    def get_default_time_format(self):
        """Get default time format."""
        return self._time_format
        
    def get_default_datetime_format(self):
        """Get default datetime format."""
        return self._datetime_format
        
    def set_1904(self, dsys=None):
        """Set the date system: 0 = 1900 (the default), 1 = 1904"""
        if dsys is not None:
            self._1904 = dsys
        else:
            self._1904 = 1

    def get_1904(self):
        """Return the date system: 0 = 1900, 1 = 1904"""
        return self._1904

    def set_custom_color(self, index, red, green=0, blue=0):
        """Change the RGB components of the elements in the colour palette."""
        if isinstance(red, str):
            # Match a HTML #xxyyzz style parameter
            res = re.match(r"^#(\w\w)(\w\w)(\w\w)", red)
            if (res):
                red, green, blue = [int(cn, 16) for cn in res.groups([1, 2, 3])]
        palette = self._palette
        # Check that the colour index is the right range
        if (index < 8 or index > 64):
            raise "Color index index outside range: 8 <= index <= 64"
        # Check that the colour components are in the right range
        if ( (red < 0 or red > 255) or \
                 (green < 0 or green > 255) or \
                 (blue < 0 or blue > 255) ):
            raise "Color component outside range: 0 <= color <= 255"
        index -= 8  # Adjust colour index (wingless dragonfly)
        # Set the RGB value
        palette[index] = [red, green, blue, 0x00]
        return index + 8

    def set_palette_xl97(self):
        """Sets the colour palette to the Excel 97+ default."""
        self._palette = [
            [0x00, 0x00, 0x00, 0x00],   # 8
            [0xff, 0xff, 0xff, 0x00],   # 9
            [0xff, 0x00, 0x00, 0x00],   # 10
            [0x00, 0xff, 0x00, 0x00],   # 11
            [0x00, 0x00, 0xff, 0x00],   # 12
            [0xff, 0xff, 0x00, 0x00],   # 13
            [0xff, 0x00, 0xff, 0x00],   # 14
            [0x00, 0xff, 0xff, 0x00],   # 15
            [0x80, 0x00, 0x00, 0x00],   # 16
            [0x00, 0x80, 0x00, 0x00],   # 17
            [0x00, 0x00, 0x80, 0x00],   # 18
            [0x80, 0x80, 0x00, 0x00],   # 19
            [0x80, 0x00, 0x80, 0x00],   # 20
            [0x00, 0x80, 0x80, 0x00],   # 21
            [0xc0, 0xc0, 0xc0, 0x00],   # 22
            [0x80, 0x80, 0x80, 0x00],   # 23
            [0x99, 0x99, 0xff, 0x00],   # 24
            [0x99, 0x33, 0x66, 0x00],   # 25
            [0xff, 0xff, 0xcc, 0x00],   # 26
            [0xcc, 0xff, 0xff, 0x00],   # 27
            [0x66, 0x00, 0x66, 0x00],   # 28
            [0xff, 0x80, 0x80, 0x00],   # 29
            [0x00, 0x66, 0xcc, 0x00],   # 30
            [0xcc, 0xcc, 0xff, 0x00],   # 31
            [0x00, 0x00, 0x80, 0x00],   # 32
            [0xff, 0x00, 0xff, 0x00],   # 33
            [0xff, 0xff, 0x00, 0x00],   # 34
            [0x00, 0xff, 0xff, 0x00],   # 35
            [0x80, 0x00, 0x80, 0x00],   # 36
            [0x80, 0x00, 0x00, 0x00],   # 37
            [0x00, 0x80, 0x80, 0x00],   # 38
            [0x00, 0x00, 0xff, 0x00],   # 39
            [0x00, 0xcc, 0xff, 0x00],   # 40
            [0xcc, 0xff, 0xff, 0x00],   # 41
            [0xcc, 0xff, 0xcc, 0x00],   # 42
            [0xff, 0xff, 0x99, 0x00],   # 43
            [0x99, 0xcc, 0xff, 0x00],   # 44
            [0xff, 0x99, 0xcc, 0x00],   # 45
            [0xcc, 0x99, 0xff, 0x00],   # 46
            [0xff, 0xcc, 0x99, 0x00],   # 47
            [0x33, 0x66, 0xff, 0x00],   # 48
            [0x33, 0xcc, 0xcc, 0x00],   # 49
            [0x99, 0xcc, 0x00, 0x00],   # 50
            [0xff, 0xcc, 0x00, 0x00],   # 51
            [0xff, 0x99, 0x00, 0x00],   # 52
            [0xff, 0x66, 0x00, 0x00],   # 53
            [0x66, 0x66, 0x99, 0x00],   # 54
            [0x96, 0x96, 0x96, 0x00],   # 55
            [0x00, 0x33, 0x66, 0x00],   # 56
            [0x33, 0x99, 0x66, 0x00],   # 57
            [0x00, 0x33, 0x00, 0x00],   # 58
            [0x33, 0x33, 0x00, 0x00],   # 59
            [0x99, 0x33, 0x00, 0x00],   # 60
            [0x99, 0x33, 0x66, 0x00],   # 61
            [0x33, 0x33, 0x99, 0x00],   # 62
            [0x33, 0x33, 0x33, 0x00],   # 63
        ]
        return 0

    def set_tempdir(self, tempdir):
        """Change the default temp directory used by _initialize() in Worksheet.pm."""
        raise NotImplementedError("use 'import tempfile; tempfile.tempdir=<tempdir>'")

        
    def set_codepage(self, codepage=1):
        """See also the _store_codepage method. This is used to store the code page, i.e.
        the character set used in the workbook.

        codepage = 1 - ANSI, MS Windows
        codepage = 2 - Apple Macintosh

        """
        if codepage == 1: codepage = 0x04E4
        if codepage == 2: codepage = 0x8000
        self._codepage = codepage

    def _store_workbook(self):
        """Assemble worksheets into a workbook and send the BIFF data to an OLE
        storage.

        """
        # Ensure that at least one worksheet has been selected.
        if (self._activesheet == 0) and (len(self._worksheets) > 0):
            self._worksheets[0]._selected = True
        # Calculate the number of selected worksheet tabs and call the finalization
        # methods for each worksheet
        for sheet in self._worksheets:
            if sheet._selected:
                self._selected += 1
            sheet._close(self._sheetnames)
        # Add Workbook globals
        self._store_bof(0x0005)
        self._store_codepage()
        self._store_window1()
        self._store_1904()
        self._store_all_fonts()
        self._store_all_num_formats()
        self._store_all_xfs()
        self._store_all_styles()
        self._store_palette()

        # Calculate the offsets required by the BOUNDSHEET records
        self._calc_sheet_offsets()

        # Add BOUNDSHEET records. For BIFF 7+ TODO ....
        for sheet in self._worksheets:
            if isinstance(sheet._name, unicode):
                name     = sheet._name.encode("utf-16-le")
                encoding = 0x1
            else:
                name     = sheet._name
                encoding = 0
            self._store_boundsheet(name,
                                   sheet._offset,
                                   encoding)

        # NOTE: If any records are added between here and EOF the
        # _calc_sheet_offsets() should be updated to include the new length.
        if self._ext_ref_count:
            self._store_supbook()
            self._store_externsheet()
            self._store_names()

        self._store_shared_strings()

        # End Workbook globals
        self._store_eof()

        # Store the workbook in an OLE container
        self._store_OLE_file()           
        
    def _store_OLE_file(self):
        """Store the workbook in an OLE container if the total size of the workbook data
        is less than ~ 7MB.

        """
        if not self._big and self._biffsize <= 7087104:
            # Write Worksheet data if data size <~ 7MB            
            ole = OLEWriter(self._file)
        else:
            ole = OLEWriterBig(self._file)
        ole.set_size(self._biffsize)
        ole.write_header()

        # This code chunk should to be redesigned (probably w/ get_data())
        # to look more pythonish (i.e. functional not imperative). AM
        tmp = self.get_data()
        while tmp != "":
            ole.write(tmp)
            tmp = self.get_data()

        for sheet in self._worksheets:
            while 1:
                tmp = sheet.get_data()
                if not tmp: break
                ole.write(tmp)
        ole.close()
            
    def _calc_sheet_offsets(self):
        """Calculate Worksheet BOF offsets for use in the BOUNDSHEET records."""
        BOF = 12
        EOF = 4
        offset = self._datasize

        # Add the length of the SST and associated CONTINUEs
        offset += self._calculate_shared_string_sizes()

        # Add the length of the SUPBOOK, EXTERNSHEET and NAME records
        offset += self._calculate_extern_sizes()

        for sheet in self._worksheets:
            offset += BOF + len(sheet._name)
        offset += EOF
        for sheet in self._worksheets:
            sheet._offset = offset
            offset += sheet._datasize
        self._biffsize = offset

    def _store_all_fonts(self):
        """Store the Excel FONT records."""
        # _tmp_format is added by new(). We use this to write the default XF's
        format = self._tmp_format
        font = format.get_font()
        # Note: Fonts are 0-indexed. According to the SDK there is no index 4,
        # so the following fonts are 0, 1, 2, 3, 5
        for i in xrange(1, 6):
            self._append(font)
        # Iterate through the XF objects and write a FONT record if it isn't the
        # same as the default FONT and if it hasn't already been used.
        fonts = {}
        index = 6                   # The first user defined FONT
        key = format.get_font_key() # The default font from _tmp_format
        fonts[key] = 0              # Index of the default font
        for format in self._formats:
            key = format.get_font_key()
            if fonts.has_key(key):
                # FONT has already been used
                format._font_index = fonts[key]
            else:
                # Add a new FONT record
                fonts[key] = index
                format._font_index = index
                index += 1
                font = format.get_font()
                self._append(font)

    def _store_all_num_formats(self):
        """Store user defined numerical formats i.e. FORMAT records."""
        # Leaning num_format syndrome
        num_formats = {}
        num_formats_ = []
        index = 164 # User defined FORMAT records start from 0xA4
        # Iterate through the XF objects and write a FORMAT record if it isn't a
        # built-in format type and if the FORMAT string hasn't already been used.
        for format in self._formats:
            num_format = format._num_format
            #encoding   = format._num_format_enc

            # Check if num_format is an index to a built-in format.
            # Also check for a string of zeros, which is a valid format string
            # but would evaluate to zero.
            if not re.match(r"^0+\d", num_format):
                if re.match(r"^\d+$", num_format):
                    continue
            if num_formats.has_key(num_format):
                # FORMAT has already been used
                format._num_format = num_formats[num_format]
            else:
                # Add a new FORMAT
                num_formats[num_format] = index
                format._num_format = index
                self._store_num_format(num_format, index)
                index+=1

    def _store_all_xfs(self):
        """Write all XF records.

        """
        # _tmp_format is added by new(). We use this to write the default XF's
        # The default font index is 0
        format = self._tmp_format
        for i in xrange(0, 15):
            xf = format.get_xf('style')     # Style XF
            self._append(xf)
        xf = format.get_xf('cell')          # Cell XF
        self._append(xf)
        # User defined XFs
        for format in self._formats:
            xf = format.get_xf('cell')
            self._append(xf)

    def _store_all_styles(self):
        """Write all STYLE records."""
        self._store_style()

    def _store_names(self):
        """Write the NAME record to define the print area and the repeat rows and cols."""
        index = 0

        # Create the print area NAME records
        for worksheet in self._worksheets:

            key = "%d:%d" % (index,index)
            ref = self._ext_refs[key]
            index += 1

            # Write a Name record if the print area has been defined
            if (worksheet._print_rowmin is not None):
                self._store_name_short(
                    worksheet._index,
                    0x06, # NAME type
                    ref,
                    worksheet._print_rowmin,
                    worksheet._print_rowmax,
                    worksheet._print_colmin,
                    worksheet._print_colmax)

        index = 0
        # Create the print title NAME records
        for worksheet in self._worksheets:
            rowmin = worksheet._title_rowmin
            rowmax = worksheet._title_rowmax
            colmin = worksheet._title_colmin
            colmax = worksheet._title_colmax
            key    = "%d:%d" % (index, index)
            ref    = self._ext_refs[key]
            index += 1

            # Determine if row + col, row, col or nothing has been defined
            # and write the appropriate record
            if ((rowmin is not None) and (colmin is not None)):
                # Row and column titles have been defined.
                # Row title has been defined.
                self._store_name_long(
                    worksheet._index,
                    0x07, # NAME type
                    ref,
                    rowmin,
                    rowmax,
                    colmin,
                    colmax)
            elif (rowmin is not None):
                # Row title has been defined.
                self._store_name_short(
                    worksheet._index,
                    0x07, # NAME type
                    ref,
                    rowmin,
                    rowmax,
                    0x00,
                    0xff)
            elif (colmin is not None):
                # Column title has been defined.
                self._store_name_short(
                    worksheet._index,
                    0x07, # NAME type
                    ref,
                    0x0000,
                    0xffff,
                    colmin,
                    colmax)
            else:
                # Nothing left to do
                pass

    ################################################################################
    #
    # BIFF RECORDS
    #
    def _store_window1(self):
        """Write Excel BIFF WINDOW1 record."""
        record = 0x003D                 # Record identifier
        length = 0x0012                 # Number of bytes to follow
        xWn = 0x0000                    # Horizontal position of window
        yWn = 0x0000                    # Vertical position of window
        dxWn = 0x25BC                   # Width of window
        dyWn = 0x1572                   # Height of window
        grbit = 0x0038                  # Option flags
        ctabsel = self._selected        # Number of workbook tabs selected
        wTabRatio = 0x0258              # Tab to scrollbar ratio
        itabFirst = self._firstsheet    # 1st displayed worksheet
        itabCur = self._activesheet     # Active worksheet
        header = pack("<HH", record, length)
        data = pack("<HHHHHHHHH", xWn, yWn, dxWn, dyWn,
                                  grbit,
                                  itabCur, itabFirst,
                                  ctabsel, wTabRatio)
        self._append(header, data)

    def _store_boundsheet(self, sheetname, offset, encoding):
        """Writes Excel BIFF BOUNDSHEET record.

        sheetname - Worksheet name
        offset    - Location of worksheet BOF

        """
        record = 0x0085                # Record identifier
        length = 0x08 + len(sheetname) # Number of bytes to follow
        grbit  = 0x0000                # Sheet identifier
        cch    = len(sheetname)        # Length of sheet name

        if encoding:
            cch /= 2
        # our sheetname is in UTF-16-LE already

        header = pack("<HH", record, length)
        data = pack("<LHBB", offset, grbit, cch, encoding) #"VvCC"
        self._append(header, data, sheetname)

    def _store_style(self):
        """Write Excel BIFF STYLE records."""
        record = 0x0293  # Record identifier
        length = 0x0004  # Bytes to follow
        ixfe = 0x8000    # Index to style XF
        BuiltIn = 0x00   # Built-in style
        iLevel = 0xff    # Outline style level
        header = pack("<HH",  record, length)
        data = pack("<HBB", ixfe, BuiltIn, iLevel)
        self._append(header, data)

    def _store_num_format(self, format, ifmt):
        """Writes Excel FORMAT record for non "built-in" numerical formats.

        format - Custom format string
        ifmt   - Format index code

        """
        record   = 0x041E                 # Record identifier

        cch      = len(format)            # Length of format string (in chars)

        if isinstance(format, unicode):
            format = format.encode("utf-16-le")
            encoding = 1
        else:
            encoding = 0

        # AM NOTE: perl version (as of 2.04 also handles Euro sign in _ascii_
        # strings specifically. This is locale-dependent and error prone
        # however so I won't do the same in python version. Those who wish
        # euro sign should use unicode

        length   = 0x05 + len(format)     # Number of bytes to follow

        header = pack("<HH", record, length)     # vv
        data = pack("<HHB", ifmt, cch, encoding) # vvC

        self._append(header, data, format)

    def _store_1904(self):
        """Write Excel 1904 record to indicate the date system in use."""
        record = 0x0022         # Record identifier
        length = 0x0002         # Bytes to follow
        f1904 = self._1904 # Flag for 1904 date system
        header = pack("<HH", record, length)
        data = pack("<H", f1904)
        self._append(header, data)

    def _store_supbook(self):
        """Write BIFF record SUPBOOK to indicate that the workbook contains external
        references, in our case, formula, print area and print title refs.
        """
        record = 0x01AE               # Record identifier
        length = 0x0004               # Number of bytes to follow

        ctabs       = len(self._worksheets)  # Number of worksheets
        StVirtPath  = 0x0401                 # Encoded workbook filename
        
        header      = pack("<HH", record, length)     # vv
        data        = pack("<HH", ctabs, StVirtPath)  # vv

        self._append(header, data)

    def _store_externsheet(self):
        """Writes the Excel BIFF EXTERNSHEET record. These references are used by
        formulas. NAME record is required to define the print area and the repeat
        rows and columns.
        """
        record = 0x0017               # Record identifier
        length = 0                    # Number of bytes to follow

        ext_refs = self._ext_refs.keys()
        from utilites import num_cmp  
        ext_refs.sort(lambda a,b: num_cmp(self._ext_refs[a],self._ext_refs[b]))
        
        # Change the external refs from stringified "1:1" to [1, 1]
        ext_refs = map(lambda m: m.split(":"), ext_refs)

        cxti        = len(ext_refs)         # Number of Excel XTI structures
        rgxti       = []                    # Array of XTI structures

        # Write the XTI structs
        for ext_ref in ext_refs:
            rgxti.append(pack("<HHH", 0, ext_ref[0], ext_ref[1]))
        rgxti  = "".join(rgxti)
        
        data   = pack("<H", cxti) + rgxti
        header = pack("<HH", record, len(data))

        self._append(header, data)

    def _store_name_short(self, index, type, ext_ref, rowmin, rowmax, colmin, colmax):
        """Store the NAME record in the short format that is used for storing the print
        area, repeat rows only and repeat columns only.

        index  - Sheet index
        type   -

        rowmin - Start row
        rowmax - End row
        colmin - Start column
        colmax - end column

        """
        record         = 0x0018       # Record identifier
        length         = 0x001b       # Number of bytes to follow
        grbit          = 0x0020       # Option flags
        chKey          = 0x00         # Keyboard shortcut
        cch            = 0x01         # Length of text name
        cce            = 0x000b       # Length of text definition
        unknown01      = 0x0000
        ixals          = index +1     # Sheet index
        unknown02      = 0x00
        cchCustMenu    = 0x00         # Length of cust menu text
        cchDescription = 0x00         # Length of description text
        cchHelptopic   = 0x00         # Length of help topic text
        cchStatustext  = 0x00         # Length of status bar text
        rgch           = type         # Built-in name type
        unknown03      = 0x3b         #

        header = pack("<HH", record, length)
        # A little optimization :-)
        data   = [pack("<H", grbit)]
        data  += pack("<B", chKey)
        data  += pack("<B", cch)
        data  += pack("<H", cce)
        data  += pack("<H", unknown01)
        data  += pack("<H", ixals)
        data  += pack("<H", unknown02)
        data  += pack("<B", cchCustMenu)
        data  += pack("<B", cchDescription)
        data  += pack("<B", cchHelptopic)
        data  += pack("<B", cchStatustext)
        data  += pack("<B", rgch)
        data  += pack("<B", unknown03)
        data  += pack("<H", ext_ref)

        data  += pack("<H", rowmin)
        data  += pack("<H", rowmax)
        data  += pack("<H", colmin)
        data  += pack("<H", colmax)

        data   = "".join(data)

        self._append(header, data)

    def _store_name_long(self, index, type, ext_ref, rowmin, rowmax, commin, commax):
        """Store the NAME record in the long format that is used for storing the repeat
        rows and columns when both are specified. This share a lot of code with
        _store_name_short() but we use a separate method to keep the code clean.
        Code abstraction for reuse can be carried too far, and I should know. -)

        index   -  Sheet index
        type

        rowmin  -  Start row
        rowmax  -  End row
        colmin  -  Start column
        colmax  -  end column

        """
        record         = 0x0018         # Record identifier
        length         = 0x002a         # Number of bytes to follow

        grbit          = 0x0020         # Option flags
        chKey          = 0x00           # Keyboard shortcut
        cch            = 0x01           # Length of text name
        cce            = 0x001a         # Length of text definition
        unknown01      = 0x0000         #
        ixals          = index + 1      # Sheet index
        unknown02      = 0x00           #
        cchCustMenu    = 0x00           # Length of cust menu text
        cchDescription = 0x00           # Length of description text
        cchHelptopic   = 0x00           # Length of help topic text
        cchStatustext  = 0x00           # Length of status bar text
        rgch           = type           # Built-in name type

        unknown03      = 0x29
        unknown04      = 0x0017
        unknown05      = 0x3b

        header = pack("<HH",  record, length)

        # A little optimization :-)
        data  = [pack("<H", grbit)]
        data += pack("<B", chKey)
        data += pack("<B", cch)
        data += pack("<H", cce)
        data += pack("<H", unknown01)
        data += pack("<H", ixals)
        data += pack("<B", unknown02)
        data += pack("<B", cchCustMenu)
        data += pack("<B", cchDescription)
        data += pack("<B", cchHelptopic)
        data += pack("<B", cchStatustext)
        data += pack("<B", rgch)

        # Column definition
        data += pack("<B", unknown03)
        data += pack("<H", unknown04)
        data += pack("<B", unknown05)
        data += pack("<H", ext_ref)
        data += pack("<H", 0x0000)
        data += pack("<H", 0xffff)

        data += pack("<H", colmin)
        data += pack("<H", colmax)

        # Row definition
        data += pack("<B", unknown05)
        data += pack("<H", ext_ref)
        data += pack("<H", rowmin)
        data += pack("<H", rowmax)
        data += pack("<H", 0x00)
        data += pack("<H", 0xff)

        # End of data
        data += pack("<B", 0x10)

        data = "".join(data)
        self._append(header, data)

    def _store_palette(self):
        """Stores the PALETTE biff record."""
        aref = self._palette
        record = 0x0092           # Record identifier
        length = 2 + 4 * len(aref)   # Number of bytes to follow
        ccv = len(aref)   # Number of RGB values to follow
        # Pack the RGB data
        data = ""
        for pal in aref:
            data += pack("<BBBB", *pal)
        header = pack("<HHH", record, length, ccv)
        self._append(header, data)

    def _store_codepage(self):
        """Stores the CODEPAGE biff record."""
        record = 0x0042               # Record identifier
        length = 0x0002               # Number of bytes to follow
        cv = self._codepage   # The code page
        header = pack("<HH", record, length)
        data = pack("<H", cv)
        self._append(header, data)

    def _calculate_extern_sizes(self):
        """We need to calculate the space required by the SUPBOOK, EXTERNSHEET and NAME
        records so that it can be added to the BOUNDSHEET offsets.
        """
        ext_refs      = self._parser.get_ext_sheets()
        ext_ref_count = len(ext_refs)
        length        = 0
        index         = 0

        for worksheet in self._worksheets:
            rowmin = worksheet._title_rowmin
            colmin = worksheet._title_colmin

            key    = "%d:%d" % (index, index)
            index += 1

            # Print area NAME records
            if worksheet._print_rowmin != None:
                if not ext_refs.has_key(key):
                    ext_refs[key] = ext_ref_count
                    ext_ref_count += 1
                length += 31

            # Print title  NAME records
            if rowmin != None and colmin != None:
                if not ext_refs.has_key(key):
                    ext_refs[key] = ext_ref_count
                    ext_ref_count += 1

                length += 46
            elif rowmin != None or colmin != None:
                if not ext_refs.has_key(key):
                    ext_refs[key] = ext_ref_count
                    ext_ref_count += 1

                length += 31
            else: pass # TODO

        self._ext_ref_count = ext_ref_count

        # If there are no external refs then we don't write, SUPBOOK, EXTERNSHEET
        # and NAME. Therefore the length is 0.

        if ext_ref_count == 0:
            return 0

        # The SUPBOOK record is 8 bytes
        length += 8

        # The EXTERNSHEET record is 6 bytes + 6 bytes for each external ref
        length += 6 * (1 + ext_ref_count)

        return length
        
    def _calculate_shared_string_sizes(self):
        """Handling of the SST continue blocks is complicated by the need to include an
        additional continuation byte depending on whether the string is split between
        blocks or whether it starts at the beginning of the block. (There are also
        additional complications that will arise later when/if Rich Strings are
        supported). As such we cannot use the simple CONTINUE mechanism provided by
        the _add_continue() method in BIFFwriter.pm. Thus we have to make two passes
        through the strings data. The first is to calculate the required block sizes
        and the second, in _store_shared_strings(), is to write the actual strings.
        The first pass through the data is also used to calculate the size of the SST
        and CONTINUE records for use in setting the BOUNDSHEET record offsets. The
        downside of this is that the same algorithm repeated in _store_shared_strings.
        """

        # Create an array of predefined size...
        strings = map(lambda _:0, xrange(self._str_unique))

        for (key, value) in self._str_table.iteritems():
            strings[value] = key

        # The SST data could be very large, free some memory (maybe).
        self._str_table = {}
        self._str_array = strings

        # Iterate through the strings to calculate the CONTINUE block sizes.
        #
        # The SST blocks requires a specialised CONTINUE block, so we have to
        # ensure that the maximum data block size is less than the limit used by
        # _add_continue() in BIFFwriter.pm. For simplicity we use the same size
        # for the SST and CONTINUE records:
        #   8228 : Maximum Excel97 block size
        #     -4 : Length of block header
        #     -8 : Length of additional SST header information
        #     -8 : Arbitrary number to keep within _add_continue() limit
        # = 8208
        #
        continue_limit = 8208
        block_length   = 0
        written        = 0
        block_sizes    = []
        cont           = 0        # continue in Workbook.pm

        for string in strings:
            string_length = len(string)
            # AM TODO
            # almost straight port of perl code
            # my $encoding  = unpack "xx C", $string;
            # Actually the python port requires a more sophisticated
            # approach to string intermediate storage and processing
            # because of 'natural' unicode support in python
            encoding = unpack("B", string[2])[0]
            split_string = False

            # Block length is the total length of the strings that will be
            # written out in a single SST or CONTINUE block.
            #
            block_length += string_length


            # We can write the string if it doesn't cross a CONTINUE boundary
            if block_length < continue_limit:
                written += string_length
                continue

            # Deal with the cases where the next string to be written will exceed
            # the CONTINUE boundary. If the string is very long it may need to be
            # written in more than one CONTINUE record.
            #
            while block_length >= continue_limit:

                # We need to avoid the case where a string is continued in the first
                # n bytes that contain the string header information.
                #
                header_length   = 3; # Min string + header size -1
                space_remaining = continue_limit - written -  cont


                # Unicode data should only be split on char (2 byte) boundaries.
                # Therefore, in some cases we need to reduce the amount of available
                # space by 1 byte to ensure the correct alignment.
                align = 0

                # Only applies to Unicode strings
                if encoding == 1:
                    # Min string + header size -1
                    header_length = 4

                    if space_remaining > header_length:
                        # String contains 3 byte header => split on odd boundary
                        # AM TODO: some really strange code... Ask John McNamara if it's correct actually
                        if not split_string and (space_remaining % 2) != 1:
                            space_remaining -= 1
                            align = 1
                        # Split section without header => split on even boundary
                        elif split_string and (space_remaining % 2) == 1:
                            space_remaining -= 1
                            align = 1

                        split_string = True

                if space_remaining > header_length:
                    # Write as much as possible of the string in the current block
                    written      += space_remaining

                    # Reduce the current block length by the amount written
                    block_length -= continue_limit - cont - align

                    # Store the max size for this block
                    block_sizes.append(continue_limit - align)

                    # If the current string was split then the next CONTINUE block
                    # should have the string continue flag (grbit) set unless the
                    # split string fits exactly into the remaining space.
                    #
                    if block_length > 0:
                        cont = 1
                    else:
                        cont = 0

                else:
                    # Store the max size for this block
                    block_sizes.append(written + cont)

                    # Not enough space to start the string in the current block
                    block_length -= continue_limit - space_remaining - cont
                    cont = 0

                # If the string (or substr) is small enough we can write it in the
                # new CONTINUE block. Else, go through the loop again to write it in
                # one or more CONTINUE blocks
                #
                if block_length < continue_limit:
                    written = block_length
                else:
                    written = 0
        # Store the max size for the last block unless it is empty
        if written + cont != 0:
            block_sizes.append(written+cont)

        self._str_block_sizes = block_sizes
        # AM NOTE: Avoid further modifications of block_sizes to prevent
        # necessity to create additional copy of the array for
        # self._str_block_sizes


        # Calculate the total length of the SST and associated CONTINUEs (if any).
        # The SST record will have a length even if it contains no strings.
        # This length is required to set the offsets in the BOUNDSHEET records since
        # they must be written before the SST records
        #
        length = 12
        if len(block_sizes): # SST
            length += block_sizes[0]
        for i in xrange(1, len(block_sizes)):
            length += 5 + block_sizes[i]

        return length

    def _store_shared_strings(self):
        """Write all of the workbooks strings into an indexed array.

        See the comments in _calculate_shared_string_sizes() for more information.

        The Excel documentation says that the SST record should be followed by an
        EXTSST record. The EXTSST record is a hash table that is used to optimise
        access to SST. However, despite the documentation it doesn't seem to be
        required so we will ignore it.
        """

        record              = 0x00FC   # Record identifier
        length              = 0x0008   # Number of bytes to follow
        total               = 0x0000

        # Iterate through the strings to calculate the CONTINUE block sizes
        continue_limit = 8208
        block_length   = 0
        written        = 0
        cont           = 0        # continue in Workbook.pm

        # The SST and CONTINUE block sizes have been pre-calculated by
        # _calculate_shared_string_sizes()
        block_sizes    = self._str_block_sizes


        # The SST record is required even if it contains no strings. Thus we will
        # always have a length
        #
        if len(block_sizes):
            length = 8 + block_sizes.pop(0)
        else: # No strings
            length = 8

        # Write the SST block header information
        header = pack("<HH", record, length)   # vv
        data   = pack("<LL", self._str_total, self._str_unique)
        self._append(header, data)


        # Iterate through the strings and write them out
        for string in self._str_array:
            string_length = len(string)
            # AM TODO
            # almost straight port of perl code
            # my $encoding  = unpack "xx C", $string;
            # Actually the python port requires a more sophisticated
            # approach to string intermediate storage and processing
            # because of 'natural' unicode support in python
            encoding     = unpack("B", string[2])[0]
            split_string = False


            # Block length is the total length of the strings that will be
            # written out in a single SST or CONTINUE block.
            #
            block_length += string_length


            # We can write the string if it doesn't cross a CONTINUE boundary
            if block_length < continue_limit:
                self._append(string)
                written += string_length
                continue

            # Deal with the cases where the next string to be written will exceed
            # the CONTINUE boundary. If the string is very long it may need to be
            # written in more than one CONTINUE record.
            #

            while block_length >= continue_limit:

                # We need to avoid the case where a string is continued in the first
                # n bytes that contain the string header information.
                #
                header_length   = 3; # Min string + header size - 1
                space_remaining = continue_limit - written - cont


                # Unicode data should only be split on char (2 byte) boundaries.
                # Therefore, in some cases we need to reduce the amount of available
                # space by 1 byte to ensure the correct alignment.
                align = 0

                # Only applies to Unicode strings
                if encoding == 1:
                    # Min string + header size - 1
                    header_length = 4

                    if space_remaining > header_length:
                        # String contains 3 byte header => split on odd boundary
                        if not split_string and space_remaining % 2 != 1:
                            space_remaining -= 1
                            align = 1
                        # Split section without header => split on even boundary
                        elif split_string and space_remaining % 2 == 1:
                            space_remaining -= 1
                            align = 1

                        split_string = True


                if space_remaining > header_length:
                    # Write as much as possible of the string in the current block
                    tmp = string[:space_remaining]
                    self._append(tmp)

                    # The remainder will be written in the next block(s)
                    string = string[space_remaining:]

                    # Reduce the current block length by the amount written
                    block_length -= continue_limit - cont - align

                    # If the current string was split then the next CONTINUE block
                    # should have the string continue flag (grbit) set unless the
                    # split string fits exactly into the remaining space.
                    #

                    if block_length > 0:
                        cont = 1
                    else:
                        cont = 0
                else:
                    # Not enough space to start the string in the current block
                    block_length -= continue_limit - space_remaining - cont
                    cont = 0

                # Write the CONTINUE block header
                if len(block_sizes):
                    record  = 0x003C
                    length  = block_sizes.pop(0)

                    header  = pack("<HH", record, length)  # vv
                    if cont:
                        header += pack("B", encoding)

                    self._append(header)

                # If the string (or substr) is small enough we can write it in the
                # new CONTINUE block. Else, go through the loop again to write it in
                # one or more CONTINUE blocks
                #
                if block_length < continue_limit:
                    self._append(string)

                    written = block_length
                else:
                    written = 0
